
from PySide6.QtWidgets import QDialog, QMainWindow, QTextEdit
from PySide6.QtGui import QActionGroup, QColor, QIcon, QTextCursor, QTextFormat
from page_settings import PageSettings
from page_help import PageHelp
from ui_pages.ui_main_window import Ui_MainWindow
from utils import FileOps, FileDialog, Client, JsonOps, Parser, Solver
from threading import Thread
from time import sleep
from os import getcwd

class MainWindow(QMainWindow):

    def __init__(self):
        self.init_parent() # Initialize parent
        self.init_ui() # Initialize graphics
        self.create_attributes() # Create attributes
        self.init_attributes() # Initialize attributes
        self.init_graphics() # Initialize graphic objects
        self.init_settings() # Initialize settings
        self.create_connections() # Create connections
        self.thread_update_start() # Start updating thread
        
    def init_parent(self):
        super().__init__()

    def init_ui(self):
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

    def create_attributes(self):
        self.file_ops = None # Service class to manage text files
        self.file_dialog = None # Service class to manage dialog windows
        self.json_ops = None # Service class to manage json files
        self.parser = None # Service class with mathematical operations
        self.client = None # Client to communicate with CNC (via CNC API)
        self.file_name = None # Name of file. Example: C:/.../myfile.ngc
        self.base_name = None # Base name of file. Example: myfile.ngc
        self.settings_file = None # Settings file
        self.status = None # Connection state (True if connected, False otherwise). Used to update graphic objects in updating thread
        self.thread_update_period = None # Period of update thread [s]
        self.thread_update = None # Updating thread
        self.thread_update_stopper = None # Stopper flag of updating thread
        self.default_header = None # Default header for new files
        self.default_body = None # Default body for new files
        self.default_footer = None # Default footer for new files
        self.default_file_text = None # Default file text produced by default triple (header, body, footer)
        self.arc_definition = None # Arc definition
        self.auto_push = None # Auto push
        self.point1_x = None # Value X of first intermediate point of arc
        self.point1_y = None # Value Y of first intermediate point of arc
        self.point1_z = None # Value Z of first intermediate point of arc
        self.point2_x = None # Value X of second intermediate point of arc
        self.point2_y = None # Value Y of second intermediate point of arc
        self.point2_z = None # Value Z of second intermediate point of arc
        self.centre_x = None # Value X of centre of arc
        self.centre_y = None # Value Y of centre of arc
        self.centre_z = None # Value Z of centre of arc

    def init_attributes(self):
        self.file_ops = FileOps()
        self.file_dialog = FileDialog()
        self.json_ops = JsonOps()
        self.parser = Parser()
        self.client = Client()
        self.file_name = None
        self.base_name = None
        self.settings_file = getcwd() + "\settings.json"
        self.status = False
        self.thread_update_period = 0.1
        self.thread_update = Thread(target=self.thread_update_run)
        self.thread_update_stopper = False
        self.default_header = "G17 G21 G40 G49 G67 G80 G90 G90.1 G94 G54" + "\n" + "F1000"
        self.default_body = ""
        self.default_footer = "M2"
        self.default_file_text = self.parser.merge(self.default_header, self.default_body, self.default_footer)
        self.arc_definition = "intermediate"
        self.auto_push = "disabled"
        self.point1_x = 0
        self.point1_y = 0
        self.point1_z = 0
        self.point2_x = None
        self.point2_y = None
        self.point2_z = None
        self.centre_x = None
        self.centre_y = None
        self.centre_z = None

    def init_graphics(self):
        self.setWindowTitle("Untitled - Jog Encoder")
        self.ui.plainTextEditHeader.setPlainText(self.default_header)
        self.ui.plainTextEditBody.setPlainText(self.default_body)
        self.ui.plainTextEditFooter.setPlainText(self.default_footer)
        self.ui.labelPositionValueX.setText("-")
        self.ui.labelPositionValueY.setText("-")
        self.ui.labelPositionValueZ.setText("-")
        self.ui.labelPoint1ValueX.setText(str(self.point1_x))
        self.ui.labelPoint1ValueY.setText(str(self.point1_y))
        self.ui.labelPoint1ValueZ.setText(str(self.point1_z))
        self.ui.labelPoint2ValueX.setText("-")
        self.ui.labelPoint2ValueY.setText("-")
        self.ui.labelPoint2ValueZ.setText("-")
        self.ui.labelCentreValueX.setText("-")
        self.ui.labelCentreValueY.setText("-")
        self.ui.labelCentreValueZ.setText("-")
        self.ui.labelStatus.setText("Not connected")
        arc_definition_group = QActionGroup(self)
        arc_definition_group.addAction(self.ui.actionArcDefinitionCentre)
        arc_definition_group.addAction(self.ui.actionArcDefinitionIntermediatePoint)
        arc_definition_group.setExclusive(True)
        self.ui.actionArcDefinitionIntermediatePoint.setChecked(True)
        auto_push_group = QActionGroup(self)
        auto_push_group.addAction(self.ui.actionAutoPushEnabled)
        auto_push_group.addAction(self.ui.actionAutoPushDisabled)
        auto_push_group.setExclusive(True)
        self.ui.actionAutoPushDisabled.setChecked(True)
        self.ui.pushButtonCentre.setVisible(False)
        self.ui.pushButtonCentreG2.setVisible(False)
        self.ui.pushButtonCentreG3.setVisible(False)
        self.ui.groupBoxCentre.setVisible(False)
        self.ui.pushButtonG0.setEnabled(False)
        self.ui.pushButtonG1.setEnabled(False)
        self.ui.pushButtonCentre.setEnabled(False)
        self.ui.pushButtonCentreG2.setEnabled(False)
        self.ui.pushButtonCentreG3.setEnabled(False)
        self.ui.pushButtonPoint.setEnabled(False)
        self.ui.pushButtonPointG2.setEnabled(False)
        self.ui.pushButtonPointG3.setEnabled(False)
        self.setWindowIcon(QIcon("icons/rosetta.ico"))
        self.ui.pushButtonG0.setIcon(QIcon("icons/g0.ico"))
        self.ui.pushButtonG1.setIcon(QIcon("icons/g1.ico"))
        self.ui.pushButtonPoint.setIcon(QIcon("icons/centre.ico"))
        self.ui.pushButtonPointG2.setIcon(QIcon("icons/point_g2.ico"))
        self.ui.pushButtonPointG3.setIcon(QIcon("icons/point_g3.ico"))
        self.ui.pushButtonCentre.setIcon(QIcon("icons/centre.ico"))
        self.ui.pushButtonCentreG2.setIcon(QIcon("icons/centre_g2.ico"))
        self.ui.pushButtonCentreG3.setIcon(QIcon("icons/centre_g3.ico"))
        self.highlight_current_line()

    def init_settings(self):
        dict = {}
        try:
            res_read_file = self.file_ops.read_file(self.settings_file)
            if (res_read_file != None):
                dict = self.json_ops.get_dict(res_read_file)
        except:
            print("Error: Failed to read \"settings.json\" file.")
        finally:
            server_host = "127.0.0.1"
            if "settings/connection/host" in dict: 
                server_host = dict["settings/connection/host"]
            server_port = 8000
            if "settings/connection/port" in dict:
                try:
                    server_port = int(dict["settings/connection/port"])
                except:
                    print("Error: Failed to load \"settings.json\" file: \"settings/connection/port\" has uncorrect syntax.")
            arc_definition = "intermediate"
            if "settings/arc_definition" in dict: 
                arc_definition = dict["settings/arc_definition"]
            auto_push = "disabled"
            if "settings/auto_push" in dict: 
                auto_push = dict["settings/auto_push"]
            self.set_server_host(server_host)
            self.set_server_port(server_port)
            if (arc_definition == "centre"):
                self.ui.actionArcDefinitionCentre.setChecked(True)
                self.on_action_arc_definition_centre_triggered()
            elif (arc_definition == "intermediate"):
                self.ui.actionArcDefinitionIntermediatePoint.setChecked(True)
                self.on_action_arc_definition_intermediate_point_triggered()
            if (auto_push == "enabled"):
                self.ui.actionAutoPushEnabled.setChecked(True)
                self.on_action_auto_push_enabled_triggered()
            elif (auto_push == "disabled"):
                self.ui.actionAutoPushDisabled.setChecked(True)
                self.on_action_auto_push_disabled_triggered()
        
    def create_connections(self):
        self.ui.actionNew.triggered.connect(self.on_action_new_triggered)
        self.ui.actionOpen.triggered.connect(self.on_action_open_triggered)
        self.ui.actionSave.triggered.connect(self.on_action_save_triggered)
        self.ui.actionSaveAs.triggered.connect(self.on_action_save_as_triggered)
        self.ui.actionQuit.triggered.connect(self.on_action_quit_triggered)
        self.ui.actionAutoPushEnabled.triggered.connect(self.on_action_auto_push_enabled_triggered)
        self.ui.actionAutoPushDisabled.triggered.connect(self.on_action_auto_push_disabled_triggered)
        self.ui.actionConnection.triggered.connect(self.on_action_connection_triggered)
        self.ui.actionImport.triggered.connect(self.on_action_import_triggered)
        self.ui.actionExport.triggered.connect(self.on_action_export_triggered)
        self.ui.actionArcDefinitionCentre.triggered.connect(self.on_action_arc_definition_centre_triggered)
        self.ui.actionArcDefinitionIntermediatePoint.triggered.connect(self.on_action_arc_definition_intermediate_point_triggered)
        self.ui.actionUsage.triggered.connect(self.on_action_usage_triggered)
        self.ui.pushButtonG0.clicked.connect(self.on_push_button_G0_clicked)
        self.ui.pushButtonG1.clicked.connect(self.on_push_button_G1_clicked)
        self.ui.pushButtonPoint.clicked.connect(self.on_push_button_point_clicked)
        self.ui.pushButtonPointG2.clicked.connect(self.on_push_button_point_G2_clicked)
        self.ui.pushButtonPointG3.clicked.connect(self.on_push_button_point_G3_clicked)
        self.ui.pushButtonCentre.clicked.connect(self.on_push_button_centre_clicked)
        self.ui.pushButtonCentreG2.clicked.connect(self.on_push_button_centre_G2_clicked)
        self.ui.pushButtonCentreG3.clicked.connect(self.on_push_button_centre_G3_clicked)
        self.ui.plainTextEditBody.cursorPositionChanged.connect(self.on_plain_text_edit_body_cursor_position_changed)

    def set_server_host(self, server_host):
        """
        Set the server host as a string.
        """
        self.client.set_server_host(server_host)

    def get_server_host(self):
        """
        Returns the server host as a string.
        """
        return(self.client.get_server_host())
    
    def set_server_port(self, server_port):
        """
        Set the server port as an integer.
        """
        self.client.set_server_port(server_port)

    def get_server_port(self):
        """
        Returns the server port as an integer.
        """
        return(self.client.get_server_port())

    def connect(self):
        """
        Connect client.
        """
        self.client.connect()

    def disconnect(self):
        """
        Disconnect client.
        """
        self.client.disconnect()

    def is_connected(self):
        """
        Returns True if client is connected, false otherwise.
        """
        return(self.client.is_connected())

    def update_data(self):
        """
        Update values read from CNC.
        """
        self.client.update()

    def get_data(self):
        """
        Return the current value of CNC data.
        """
        return(self.client.cnc_data)

    def are_axes_ok(self):
        """
        Return True if axes X and Y are enabled.
        """
        return(self.client.are_axes_ok())

    def set_cnc_gcode_text(self, text):
        """
        Set a new text for CNC gcode editor
        """
        self.client.set_gcode_text(text)

    def thread_update_start(self):
        """
        Start updating thread.
        """
        self.thread_update.start()

    def thread_update_stop(self):
        """
        Stop updating thread.
        """
        self.thread_update_stopper = True

    def thread_update_run(self):
        """
        Main function of updating thread.
        """
        while(True):
            # Check stopper
            if (self.thread_update_stopper == True):
                break
            # Update values. If not connected, try to reconnect
            if (self.is_connected() == True):
                self.update_data()
                if (self.are_axes_ok() == True):
                    pos_x = self.get_data().axes_info.program_position[0]
                    pos_y = self.get_data().axes_info.program_position[1]
                    pos_z = self.get_data().axes_info.program_position[2]
                    self.ui.labelPositionValueX.setText("{:.3f}".format(pos_x))
                    self.ui.labelPositionValueY.setText("{:.3f}".format(pos_y))
                    self.ui.labelPositionValueZ.setText("{:.3f}".format(pos_z))
                    self.ui.labelStatus.setText("Connected with " + str(self.get_server_host()) + ":" + str(self.get_server_port()))
                    if (self.status == False):
                        # Connected and previously not connected: enable buttons
                        self.ui.pushButtonG0.setEnabled(True)
                        self.ui.pushButtonG1.setEnabled(True)
                        self.ui.pushButtonCentre.setEnabled(True)
                        if (self.ui.pushButtonCentre.isChecked() == True):
                            self.ui.pushButtonCentreG2.setEnabled(True)
                            self.ui.pushButtonCentreG3.setEnabled(True)
                        else:
                            self.ui.pushButtonCentreG2.setEnabled(False)
                            self.ui.pushButtonCentreG3.setEnabled(False)
                        self.ui.pushButtonPoint.setEnabled(True)
                        if ((self.ui.pushButtonPoint.isChecked() == True)):
                            self.ui.pushButtonPointG2.setEnabled(True)
                            self.ui.pushButtonPointG3.setEnabled(True)
                        else:
                            self.ui.pushButtonPointG2.setEnabled(False)
                            self.ui.pushButtonPointG3.setEnabled(False)
                        self.status = True
                else:
                    string = "Connected with " + str(self.get_server_host()) + ":" + str(self.get_server_port()) + ". "
                    string += "Warning: axis X or Y not enabled."
                    self.ui.labelStatus.setText(string)    
            else:
                if (self.status == True):
                    # Not connected and previously connected: disable buttons
                    self.ui.pushButtonG0.setEnabled(False)
                    self.ui.pushButtonG1.setEnabled(False)
                    self.ui.pushButtonCentre.setEnabled(False)
                    self.ui.pushButtonCentreG2.setEnabled(False)
                    self.ui.pushButtonCentreG3.setEnabled(False)
                    self.ui.pushButtonPoint.setEnabled(False)
                    self.ui.pushButtonPointG2.setEnabled(False)
                    self.ui.pushButtonPointG3.setEnabled(False)
                    self.status = False
                self.ui.labelPositionValueX.setText("-")
                self.ui.labelPositionValueY.setText("-")
                self.ui.labelPositionValueZ.setText("-")
                self.ui.labelStatus.setText("Not connected")
                self.connect()
            sleep(self.thread_update_period)
            
    def is_text_modified(self):
        """
        Returns True if any one between header, body and footer is changed, False otherwise.
        """
        is_header_modified = self.ui.plainTextEditHeader.document().isModified()
        is_body_modified = self.ui.plainTextEditBody.document().isModified()
        is_footer_modified = self.ui.plainTextEditFooter.document().isModified()
        is_modified = (is_header_modified or is_body_modified or is_footer_modified)
        return(is_modified)

    def get_text(self):
        """
        Returns the text obtained by merging header, body and footer.
        """
        header = self.ui.plainTextEditHeader.toPlainText()
        body = self.ui.plainTextEditBody.toPlainText()
        footer = self.ui.plainTextEditFooter.toPlainText()
        text = self.parser.merge(header, body, footer)
        return(text)

    def highlight_current_line(self):
        """
        Highlight current line of body text.
        """
        extra_selection_list = []
        extra_selection = QTextEdit.ExtraSelection()
        extra_selection.format.setBackground(QColor("#00A9FB").lighter(170))
        extra_selection.format.setProperty(QTextFormat.FullWidthSelection, True)
        extra_selection.cursor = self.ui.plainTextEditBody.textCursor()
        extra_selection.cursor.clearSelection()
        extra_selection_list.append(extra_selection)
        self.ui.plainTextEditBody.setExtraSelections(extra_selection_list)

    def on_action_new_triggered(self):
        """
        Button clicked: File -> New
        """
        new_file = False
        is_text_modified = self.is_text_modified()
        if (is_text_modified == True):
            filter = "GCode files (*.ngc)"
            text = self.get_text()
            res_sync_file = self.file_dialog.sync_file(self.file_name, filter, text)
            if (res_sync_file == True):
                new_file = True
        else:
            new_file = True
        if (new_file == True):
            self.file_name = None
            self.base_name = None
            self.ui.plainTextEditHeader.setPlainText(self.default_header)
            self.ui.plainTextEditBody.setPlainText(self.default_body)
            self.ui.plainTextEditFooter.setPlainText(self.default_footer)
            self.setWindowTitle("Untitled - Jog Encoder")
    
    def on_action_open_triggered(self):
        """
        Button clicked: File -> Open
        """
        open_file = False
        is_text_modified = self.is_text_modified()
        if (is_text_modified == True):
            filter = "GCode files (*.ngc)"
            text = self.get_text()
            res_sync_file = self.file_dialogsync_file(self.file_name, filter, text)
            if (res_sync_file == True):
                open_file = True
        else:
            open_file = True
        if (open_file == True):
            filter = "GCode files (*.ngc)"
            res_open_file = self.file_dialog.open_file(filter)
            if (res_open_file != None):
                (file_name, base_name) = res_open_file
                res_read_file = self.file_ops.read_file(file_name)
                if (res_read_file != None):
                    self.file_name = file_name
                    self.base_name = base_name
                    (header, body, footer) = self.parser.split(res_read_file)
                    self.ui.plainTextEditHeader.setPlainText(header)
                    self.ui.plainTextEditBody.setPlainText(body)
                    self.ui.plainTextEditFooter.setPlainText(footer)
                    self.setWindowTitle(self.base_name + " - Jog Encoder")

    def on_action_save_triggered(self):
        """
        Button clicked: File -> Save
        """
        if (self.file_name != None):
            text = self.get_text()
            res_write_file = self.file_ops.write_file(self.file_name, text)
        else:
            filter = "GCode files (*.ngc)"
            res_save_file = self.file_dialog.save_file(filter)
            if (res_save_file != None):
                    (file_name, base_name) = res_save_file
                    text = self.get_text()
                    res_write_file = self.file_ops.write_file(file_name, text)
                    if (res_write_file != None):
                        self.file_name = file_name
                        self.base_name = base_name

    def on_action_save_as_triggered(self):
        """
        Button clicked: File -> Save as
        """
        filter = "GCode files (*.ngc)"
        res_save_file = self.file_dialog.save_file(filter)
        if (res_save_file != None):
                (file_name, base_name) = res_save_file
                header = self.ui.plainTextEditHeader.toPlainText()
                body = self.ui.plainTextEditBody.toPlainText()
                footer = self.ui.plainTextEditFooter.toPlainText()
                text = self.parser.merge(header, body, footer)
                res_write_file = self.file_ops.write_file(file_name, text)
                if (res_write_file != None):
                    self.file_name = file_name
                    self.base_name = base_name

    def on_action_quit_triggered(self):
        """
        Button clicked: File -> Quit
        """
        self.close() # Triggers closeEvent()

    def closeEvent(self, event):
        """
        Override of quit application event
        """
        quit = True
        is_text_modified = self.is_text_modified()
        if (is_text_modified == True):
            filter = "GCode files (*.ngc)"
            text = self.get_text()
            res_sync_file = self.file_dialog.sync_file(self.file_name, filter, text)
            if (res_sync_file != True):
                quit = False
        if (quit == True):
            try:
                dict = {}
                dict["settings/connection/host"] = self.get_server_host()
                dict["settings/connection/port"] = self.get_server_port()
                dict["settings/arc_definition"] = self.arc_definition
                dict["settings/auto_push"] = self.auto_push
                text = self.json_ops.get_string(dict)
                self.file_ops.write_file(self.settings_file, text)
                self.thread_update_stop()
                self.disconnect()
            except:
                pass
            event.accept()
        else:
            event.ignore()

    def on_action_connection_triggered(self):
        """
        Button clicked: Settings -> Connection
        """
        dialog = PageSettings()
        dialog.ui.lineEditHost.setText(str(self.get_server_host()))
        dialog.ui.lineEditPort.setText(str(self.get_server_port()))
        result = dialog.exec()
        if (result == QDialog.Accepted):
            self.disconnect()
            self.set_server_host(dialog.ui.lineEditHost.text())
            self.set_server_port(int(dialog.ui.lineEditPort.text()))
    
    def on_action_arc_definition_centre_triggered(self):
        """
        Button clicked: Settings -> Arc definition -> Centre
        """
        self.arc_definition = "centre"
        self.ui.pushButtonPoint.setVisible(False)
        self.ui.pushButtonPointG2.setVisible(False)
        self.ui.pushButtonPointG3.setVisible(False)
        self.ui.groupBoxPoint1.setVisible(False)
        self.ui.groupBoxPoint2.setVisible(False)
        self.ui.pushButtonCentre.setVisible(True)
        self.ui.pushButtonCentreG2.setVisible(True)
        self.ui.pushButtonCentreG3.setVisible(True)
        self.ui.groupBoxCentre.setVisible(True)

    def on_action_arc_definition_intermediate_point_triggered(self):
        """
        Button clicked: Settings -> Arc definition -> Intermediate point
        """
        self.arc_definition = "intermediate"
        self.ui.pushButtonCentre.setVisible(False)
        self.ui.pushButtonCentreG2.setVisible(False)
        self.ui.pushButtonCentreG3.setVisible(False)
        self.ui.groupBoxCentre.setVisible(False)
        self.ui.pushButtonPoint.setVisible(True)
        self.ui.pushButtonPointG2.setVisible(True)
        self.ui.pushButtonPointG3.setVisible(True)
        self.ui.groupBoxPoint1.setVisible(True)
        self.ui.groupBoxPoint2.setVisible(True)

    def on_action_auto_push_enabled_triggered(self):
        """
        Button clicked: Settings -> Auto push -> Enabled
        """
        self.auto_push = "enabled"

    def on_action_auto_push_disabled_triggered(self):
        """
        Button clicked: Settings -> Auto push -> Disabled
        """
        self.auto_push = "disabled"

    def on_action_import_triggered(self):
        """
        Button clicked: Settings -> Import
        """
        filter = "Json files (*.json)"
        res_open_file = self.file_dialog.open_file(filter)
        if (res_open_file != None):
            (file_name, base_name) = res_open_file
            res_read_file = self.file_ops.read_file(file_name)
            if (res_read_file != None):
                try:
                    dict = self.json_ops.get_dict(res_read_file)
                    server_host = dict["settings/connection/host"]
                    server_port = dict["settings/connection/port"]
                    arc_definition = dict["settings/arc_definition"]
                    auto_push = dict["settings/auto_push"]
                    self.set_server_host(server_host)
                    self.set_server_port(server_port)
                    if (arc_definition == "centre"):
                        self.ui.actionArcDefinitionCentre.setChecked(True)
                        self.on_action_arc_definition_centre_triggered()
                    elif (arc_definition == "intermediate"):
                        self.ui.actionArcDefinitionIntermediatePoint.setChecked(True)
                        self.on_action_arc_definition_intermediate_point_triggered()
                    if (auto_push == "enabled"):
                        self.ui.actionAutoPushEnabled.setChecked(True)
                        self.on_action_auto_push_enabled_triggered()
                    elif (auto_push == "disabled"):
                        self.ui.actionAutoPushDisabled.setChecked(True)
                        self.on_action_auto_push_disabled_triggered()
                except:
                    print("Error: Failed to read json file.")

    def on_action_export_triggered(self):
        """
        Button clicked: Settings -> Export
        """
        filter = "Json files (*.json)"
        res_save_file = self.file_dialog.save_file(filter)
        if (res_save_file != None):
                (file_name, base_name) = res_save_file
                dict = {}
                dict["settings/connection/host"] = self.get_server_host()
                dict["settings/connection/port"] = self.get_server_port()
                dict["settings/arc_definition"] = self.arc_definition
                dict["settings/auto_push"] = self.auto_push
                text = self.json_ops.get_string(dict)
                res_write_file = self.file_ops.write_file(file_name, text)

    def on_action_usage_triggered(self):
        """
        Button clicked: Help -> Usage
        """
        usage_window = PageHelp()
        usage_window.exec()
    
    def on_push_button_G0_clicked(self):
        """
        Button clicked: G0
        """
        pos_x = self.get_data().axes_info.program_position[0]
        pos_y = self.get_data().axes_info.program_position[1]
        pos_z = self.get_data().axes_info.program_position[2]
        self.point1_x = pos_x
        self.point1_y = pos_y
        self.point1_z = pos_z
        self.ui.labelPoint1ValueX.setText(str(self.point1_x))
        self.ui.labelPoint1ValueY.setText(str(self.point1_y))
        self.ui.labelPoint1ValueZ.setText(str(self.point1_z))
        new_line = "G0 X" + str(pos_x) + " Y" + str(pos_y) + "\n"
        text_cursor = self.ui.plainTextEditBody.textCursor()
        text_cursor.movePosition(QTextCursor.StartOfLine)
        text_cursor.insertText(new_line)
        self.ui.plainTextEditBody.ensureCursorVisible()
        if (self.auto_push == "enabled"):
            text = self.get_text()
            self.set_cnc_gcode_text(text)

    def on_push_button_G1_clicked(self):
        """
        Button clicked: G1
        """
        pos_x = self.get_data().axes_info.program_position[0]
        pos_y = self.get_data().axes_info.program_position[1]
        pos_z = self.get_data().axes_info.program_position[2]
        self.point1_x = pos_x
        self.point1_y = pos_y
        self.point1_z = pos_z
        self.ui.labelPoint1ValueX.setText(str(self.point1_x))
        self.ui.labelPoint1ValueY.setText(str(self.point1_y))
        self.ui.labelPoint1ValueZ.setText(str(self.point1_z))
        new_line = "G1 X" + str(pos_x) + " Y" + str(pos_y) + "\n"
        text_cursor = self.ui.plainTextEditBody.textCursor()
        text_cursor.movePosition(QTextCursor.StartOfLine)
        text_cursor.insertText(new_line)
        self.ui.plainTextEditBody.ensureCursorVisible()
        if (self.auto_push == "enabled"):
            text = self.get_text()
            self.set_cnc_gcode_text(text)
    
    def on_push_button_point_clicked(self):
        """
        Button clicked: point
        """
        if (self.ui.pushButtonPoint.isChecked() == True):
            self.point2_x = self.get_data().axes_info.program_position[0]
            self.point2_y = self.get_data().axes_info.program_position[1]
            self.point2_z = self.get_data().axes_info.program_position[2]
            self.ui.labelPoint2ValueX.setText(str(self.point2_x))
            self.ui.labelPoint2ValueY.setText(str(self.point2_y))
            self.ui.labelPoint2ValueZ.setText(str(self.point2_z))
            self.ui.pushButtonPointG2.setEnabled(True)
            self.ui.pushButtonPointG3.setEnabled(True)
        else:
            self.ui.pushButtonPointG2.setEnabled(False)
            self.ui.pushButtonPointG3.setEnabled(False)
            self.point2_x = None
            self.point2_y = None
            self.point2_z = None
            self.ui.labelPoint2ValueX.setText("-")
            self.ui.labelPoint2ValueY.setText("-")
            self.ui.labelPoint2ValueZ.setText("-")

    def on_push_button_point_G2_clicked(self):
        """
        Button clicked: point G2
        """
        self.ui.pushButtonPointG2.setEnabled(False)
        self.ui.pushButtonPointG3.setEnabled(False)
        self.ui.pushButtonPoint.setChecked(False)
        pos_x = self.get_data().axes_info.program_position[0]
        pos_y = self.get_data().axes_info.program_position[1]
        pos_z = self.get_data().axes_info.program_position[2]
        solver = Solver()
        point1 = (self.point1_x, self.point1_y)
        point2 = (self.point2_x, self.point2_y)
        point3 = (pos_x, pos_y)
        centre = solver.get_centre(point1, point2, point3)
        if (centre == None):
            new_line = "; Error: The three points are aligned. Arc cannot be drawn." + "\n"
        else:
            self.point1_x = pos_x
            self.point1_y = pos_y
            self.point1_z = pos_z
            self.ui.labelPoint1ValueX.setText(str(self.point1_x))
            self.ui.labelPoint1ValueY.setText(str(self.point1_y))
            self.ui.labelPoint1ValueZ.setText(str(self.point1_z))
            (centre_x, centre_y) = centre
            centre_x = round(centre_x, 4)
            centre_y = round(centre_y, 4)
            new_line = "G2 X" + str(pos_x) + " Y" + str(pos_y) + " I" + str(centre_x) + " J" + str(centre_y) + "\n"
        self.point2_x = None
        self.point2_y = None
        self.point2_z = None
        self.ui.labelPoint2ValueX.setText("-")
        self.ui.labelPoint2ValueY.setText("-")
        self.ui.labelPoint2ValueZ.setText("-")
        text_cursor = self.ui.plainTextEditBody.textCursor()
        text_cursor.movePosition(QTextCursor.StartOfLine)
        text_cursor.insertText(new_line)
        self.ui.plainTextEditBody.ensureCursorVisible()
        if (self.auto_push == "enabled"):
            text = self.get_text()  
            self.set_cnc_gcode_text(text)

    def on_push_button_point_G3_clicked(self):
        """
        Button clicked: point G3
        """
        self.ui.pushButtonPointG2.setEnabled(False)
        self.ui.pushButtonPointG3.setEnabled(False)
        self.ui.pushButtonPoint.setChecked(False)
        pos_x = self.get_data().axes_info.program_position[0]
        pos_y = self.get_data().axes_info.program_position[1]
        pos_z = self.get_data().axes_info.program_position[2]
        solver = Solver()
        point1 = (self.point1_x, self.point1_y)
        point2 = (self.point2_x, self.point2_y)
        point3 = (pos_x, pos_y)
        centre = solver.get_centre(point1, point2, point3)
        if (centre == None):
            new_line = "; Error: The three points are aligned. Arc cannot be drawn." + "\n"
        else:
            self.point1_x = pos_x
            self.point1_y = pos_y
            self.point1_z = pos_z
            self.ui.labelPoint1ValueX.setText(str(self.point1_x))
            self.ui.labelPoint1ValueY.setText(str(self.point1_y))
            self.ui.labelPoint1ValueZ.setText(str(self.point1_z))
            (centre_x, centre_y) = centre
            centre_x = round(centre_x, 4)
            centre_y = round(centre_y, 4)
            new_line = "G3 X" + str(pos_x) + " Y" + str(pos_y) + " I" + str(centre_x) + " J" + str(centre_y) + "\n"
        self.point2_x = None
        self.point2_y = None
        self.point2_z = None
        self.ui.labelPoint2ValueX.setText("-")
        self.ui.labelPoint2ValueY.setText("-")
        self.ui.labelPoint2ValueZ.setText("-")
        text_cursor = self.ui.plainTextEditBody.textCursor()
        text_cursor.movePosition(QTextCursor.StartOfLine)
        text_cursor.insertText(new_line)
        self.ui.plainTextEditBody.ensureCursorVisible()
        if (self.auto_push == "enabled"):
            text = self.get_text()
            self.set_cnc_gcode_text(text)

    def on_push_button_centre_clicked(self):
        """
        Button clicked: centre
        """
        if (self.ui.pushButtonCentre.isChecked() == True):
            self.ui.pushButtonCentreG2.setEnabled(True)
            self.ui.pushButtonCentreG3.setEnabled(True)
            self.centre_x = self.get_data().axes_info.program_position[0]
            self.centre_y = self.get_data().axes_info.program_position[1]
            self.centre_z = self.get_data().axes_info.program_position[2]
            self.ui.labelCentreValueX.setText(str(self.centre_x))
            self.ui.labelCentreValueY.setText(str(self.centre_y))
            self.ui.labelCentreValueZ.setText(str(self.centre_z))
        else:
            self.ui.pushButtonCentreG2.setEnabled(False)
            self.ui.pushButtonCentreG3.setEnabled(False)
            self.centre_x = None
            self.centre_y = None
            self.centre_z = None
            self.ui.labelCentreValueX.setText("-")
            self.ui.labelCentreValueY.setText("-")
            self.ui.labelCentreValueZ.setText("-")
            
    def on_push_button_centre_G2_clicked(self):
        """
        Button clicked: centre G2
        """
        self.ui.pushButtonCentreG2.setEnabled(False)
        self.ui.pushButtonCentreG3.setEnabled(False)
        self.ui.pushButtonCentre.setChecked(False)
        pos_x = self.get_data().axes_info.program_position[0]
        pos_y = self.get_data().axes_info.program_position[1]
        pos_z = self.get_data().axes_info.program_position[2]
        self.point1_x = pos_x
        self.point1_y = pos_y
        self.point1_z = pos_z
        self.ui.labelPoint1ValueX.setText(str(self.point1_x))
        self.ui.labelPoint1ValueY.setText(str(self.point1_y))
        self.ui.labelPoint1ValueZ.setText(str(self.point1_z))
        new_line = "G2 X" + str(pos_x) + " Y" + str(pos_y) + " I" + str(self.centre_x) + " J" + str(self.centre_y) + "\n"
        self.centre_x = None
        self.centre_y = None
        self.centre_z = None
        self.ui.labelCentreValueX.setText("-")
        self.ui.labelCentreValueY.setText("-")
        self.ui.labelCentreValueZ.setText("-")
        text_cursor = self.ui.plainTextEditBody.textCursor()
        text_cursor.movePosition(QTextCursor.StartOfLine)
        text_cursor.insertText(new_line)
        self.ui.plainTextEditBody.ensureCursorVisible()
        if (self.auto_push == "enabled"):
            text = self.get_text()
            self.set_cnc_gcode_text(text)

    def on_push_button_centre_G3_clicked(self):
        """
        Button clicked: centre G3
        """
        self.ui.pushButtonCentreG2.setEnabled(False)
        self.ui.pushButtonCentreG3.setEnabled(False)
        self.ui.pushButtonCentre.setChecked(False)
        pos_x = self.get_data().axes_info.program_position[0]
        pos_y = self.get_data().axes_info.program_position[1]
        pos_z = self.get_data().axes_info.program_position[2]
        self.point1_x = pos_x
        self.point1_y = pos_y
        self.point1_z = pos_z
        self.ui.labelPoint1ValueX.setText(str(self.point1_x))
        self.ui.labelPoint1ValueY.setText(str(self.point1_y))
        self.ui.labelPoint1ValueZ.setText(str(self.point1_z))
        new_line = "G3 X" + str(pos_x) + " Y" + str(pos_y) + " I" + str(self.centre_x) + " J" + str(self.centre_y) + "\n"
        self.centre_x = None
        self.centre_y = None
        self.centre_z = None
        self.ui.labelCentreValueX.setText("-")
        self.ui.labelCentreValueY.setText("-")
        self.ui.labelCentreValueZ.setText("-")
        text_cursor = self.ui.plainTextEditBody.textCursor()
        text_cursor.movePosition(QTextCursor.StartOfLine)
        text_cursor.insertText(new_line)
        self.ui.plainTextEditBody.ensureCursorVisible()
        if (self.auto_push == "enabled"):
            text = self.get_text()
            self.set_cnc_gcode_text(text)

    def on_plain_text_edit_body_cursor_position_changed(self):
        """
        Event triggered: changed position of body text cursor.
        Effect: highlight current line of body text.
        """
        self.highlight_current_line()