
from PySide6.QtWidgets import QFileDialog, QMessageBox
from PySide6.QtGui import QIcon
import cnc_api_client_core as cnc
from os.path import basename
import json
import numpy

class FileOps(object):
    """
    Class containing operations with files.
    """

    def read_file(self, file_name):
        """
        Returns the text of file. If failed, returns None.
        """
        try:
            file = open(file_name, "r")
            text = file.read()
            file.close()
            return(text)
        except:
            print("Error: Failed to read file: " + str(file_name) + ".")
            return(None)

    def write_file(self, file_name, text):
        """
        Write the text on file. If failed, returns None.
        """
        try:
            file = open(file_name, "w+")
            file.write(text)
            file.close()
            return(text)
        except:
            print("Error: Failed to write file: " + str(file_name) + ".")
            return(None)

    def are_equal(self, file_name, text):
        """
        Check if file text is equal to argument text.

        Returns True if they are equal, False otherwise.
        If failed, returns None.
        """
        file_text = self.read_file(file_name)
        if (file_text == None):
            return(None)
        elif (file_text == text):
            return(True)
        else:
            return(False)

class FileDialog(object):
    """
    Class containing dialogs about files.
    """

    def __init__(self):
        self.file_ops = FileOps()

    def save_changes(self):
        """
        Dialog: "Do you want to save your changes?"

        Returns True for "Yes", False for "No", None otherwise (for "Cancel" or "X").
        """
        dialog = QMessageBox()
        dialog.setWindowTitle("Jog Encoder")
        dialog.setText("Do you want to save your changes?")
        dialog.setWindowIcon(QIcon("icons/rosetta.ico"))
        dialog.setStandardButtons(QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel)
        dialog.setDefaultButton(QMessageBox.Yes)
        result = dialog.exec()
        if (result == QMessageBox.Yes):
            return(True)
        elif (result == QMessageBox.No):
            return(False)
        else:
            return(None)
    
    def save_file(self, filter):
        """
        Dialog: "Where do you want to save the file?"

        Returns the pair (file_name, base_name). If canceled, returns None.
        """
        dialog = QFileDialog()
        (file_name, _) = dialog.getSaveFileName(filter=filter, dir="*")
        if (file_name == ""):
            return(None)
        base_name = basename(file_name)
        return((file_name, base_name))

    def open_file(self, filter):
        """
        Dialog: "What is the file you want to open?"

        Returns the pair (file_name, base_name). If canceled, returns None.
        """
        dialog = QFileDialog()
        (file_name, _) = dialog.getOpenFileName(filter=filter)
        if (file_name == ""):
            return(None)
        base_name = basename(file_name)
        return((file_name, base_name))

    def sync_file(self, file_name, filter, text):
        """
        Multi Dialog
        - Dialog: "Do you want to save your changes?"
        - Dialog: "Where do you want to save the file?" (if file_name is None)
        - Write the text on the file_name

        Returns False if user canceled some operation at some point, True
        otherwise. If some operation failed, returns None.
        """
        # Save changes: "Do you want to save your changes?"
        res_save_changes = self.save_changes()
        if (res_save_changes == None):
            return(None)
        elif (res_save_changes == False):
            return(True)
        elif (res_save_changes == True):
            pass
        # Save file: "Where do you want to save the file?"
        if (file_name == None):
            res_save_file = self.save_file(filter)
            if (res_save_file == None):
                return(False)
            (file_name_new, _) = res_save_file
        else:
            file_name_new = file_name
        # Write file
        res_write_file = self.file_ops.write_file(file_name_new, text)
        if (res_write_file == None):
            return(None)
        else:
            return(True)

class JsonOps(object):
    """
    Class containing operations with json files.
    """

    def get_dict(self, json_string):
        """
        Convert a json string into a dictionary.
        """
        dict = json.loads(json_string)
        return(dict)

    def get_string(self, dict):
        """
        Convert a dictionary into a json string.
        """
        json_string = json.dumps(dict, indent=4)
        return(json_string)

class Client(object):
    """
    Class implementing a client of CNC API.

    Usage: set_server_host() -> set_server_port() ->
    -> connect() -> -> disconnect(). If URL changes,
    you have to disconnect and connect again. When
    connected, use update() to update values of data
    read from CNC.
    """

    def __init__(self):
        self.create_attributes() # Create attributes
        self.init_attributes() # Initialize attributes

    def create_attributes(self):
        self.server_host = None # Server host
        self.server_port = None # Server port
        self.cnc_api = None # Api to communicate with CNC
        self.cnc_data = None # Service class containing CNC data

    def init_attributes(self):
        del(self.cnc_api)
        del(self.cnc_data)
        self.cnc_api = cnc.CncAPIClientCore()
        self.cnc_data = cnc.CncAPIInfoContext(self.cnc_api)

    def set_server_host(self, server_host):
        self.server_host = server_host

    def get_server_host(self):
        return(self.server_host)
    
    def set_server_port(self, server_port):
        self.server_port = server_port

    def get_server_port(self):
        return(self.server_port)

    def connect(self):
        self.init_attributes() # We create a new instance of CncAPIClientCore() because if we disconnect() then we cannot connect() again, apparently.
        self.cnc_api.connect(self.server_host, self.server_port)

    def disconnect(self):
        self.cnc_api.close()

    def is_connected(self):
        return(self.cnc_api.is_connected)

    def update(self):
        self.cnc_data.update()

    def are_axes_ok(self):
        axis_x_enabled = (self.cnc_data.cnc_info.axes_mask & cnc.X_AXIS_MASK) != 0
        axis_y_enabled = (self.cnc_data.cnc_info.axes_mask & cnc.Y_AXIS_MASK) != 0
        result = axis_x_enabled and axis_y_enabled
        return(result)

    def set_gcode_text(self, text):
        self.cnc_api.program_gcode_set_text(text)
        self.cnc_api.program_analysis(cnc.ANALYSIS_MT)

class Parser(object):
    """
    Class containing translators for ngc texts.
    """

    def __init__(self):
        self.separator = "; ------------------------------"

    def merge(self, header, body, footer):
        """
        Returns the text representing the triple (header, body, footer).
        """
        separator = "\n" + self.separator + "\n"
        text = header + separator + body + separator + footer
        return(text)

    def split(self, text):
        """
        Returns the triple (header, body, footer) represented by the text.
        """
        header = ""
        body = ""
        footer = ""
        lines = text.splitlines()
        section = "header"
        section_text = ""
        first_line = True
        for line in lines:
            if (line != self.separator):
                if (first_line == True):
                    section_text += line
                    first_line = False
                else:
                    section_text += ("\n" + line)
            else:
                if (section == "header"):
                    header = section_text
                    section = "body"
                    section_text = ""
                    first_line = True
                elif (section == "body"):
                    body = section_text
                    section = "footer"
                    section_text = ""
                    first_line = True
                elif (section == "footer"):
                    footer = section_text
                    section = "other"
        if (section == "header"):
            header = section_text
        elif (section == "body"):
            body = section_text
        elif (section == "footer"):
            footer = section_text
        return((header, body, footer))

class Solver(object):
    """
    Class containing a solver that computes the centre of an arc given three points.
    """

    def __init__(self):
        self.point1_x = None
        self.point1_y = None
        self.point2_x = None
        self.point2_y = None
        self.point3_x = None
        self.point3_y = None
        self.matrix = None
        self.column = None
        self.solution = None
        self.centre = None

    def _set_points(self, point1, point2, point3):
        (self.point1_x, self.point1_y) = point1
        (self.point2_x, self.point2_y) = point2
        (self.point3_x, self.point3_y) = point3

    def _compute_matrix(self):
        line1 = [self.point1_x, self.point1_y, 1]
        line2 = [self.point2_x, self.point2_y, 1]
        line3 = [self.point3_x, self.point3_y, 1]
        self.matrix = [line1, line2, line3]

    def _compute_column(self):
        line1 = - (self.point1_x ** 2) - (self.point1_y ** 2)
        line2 = - (self.point2_x ** 2) - (self.point2_y ** 2)
        line3 = - (self.point3_x ** 2) - (self.point3_y ** 2)
        self.column = [line1, line2, line3]

    def _solve(self):
        matrix = numpy.array(self.matrix)
        column = numpy.array(self.column)
        solution_ndarray = numpy.linalg.solve(matrix, column)
        self.solution = solution_ndarray.tolist()

    def _get_centre(self):
        self.centre = [- (self.solution[0] / 2), - (self.solution[1] / 2)]

    def get_centre(self, point1, point2, point3):
        """
        Returns the centre (x, y) of the arc passing through three given points.
        Returns None if such arc does not exist (when the three points are aligned).
        """
        self._set_points(point1, point2, point3)
        self._compute_matrix()
        self._compute_column()
        try:
            self._solve()
        except:
            return(None)
        self._get_centre()
        return(self.centre)

