"""This module implements os dependent utils."""
#-------------------------------------------------------------------------------
# Name:         os_utils
#
# Purpose:      This module implements os dependent utils.
#
# Note          Tested with Python 3.9.7
#
# Requires:     None
#
# Author:       support@rosettacnc.com
#
# Created:      09/05/2022
# Copyright:    RosettaCNC (c) 2016-2022
# Licence:      RosettaCNC License 1.0 (RCNC-1.0)
# Coding Style  https://www.python.org/dev/peps/pep-0008/
#-------------------------------------------------------------------------------
# pylint: disable=R0903 -> too-few-public-methods
# pylint: disable=W0603 -> global-statement
# pylint: disable=W0612 -> unused-variable
#-------------------------------------------------------------------------------
import contextlib
import ctypes
from ctypes import wintypes
import ipaddress
import sys

# global variables
OS_IMPROVE_TIMINGS_IN_USE = None

# == BEG: public attributes
#

def is_argument_available(argument: str) -> bool:
    """
    Search for argument in command line ignoring string case.

    return      True if argument is found.
    """
    for arg in sys.argv:
        arg = arg.lower()
        pos = arg.find(argument)
        if pos == 0:
            return True
    return False

def get_argument_value_as_float(argument: str, value_separator: str = '=') -> float:
    """
    Get argument value as float from command line ignoring string case.

    return      A float value otherwise None if argument not found or found an invalid value type.
    """
    arg = get_argument_value_as_str(argument, value_separator)
    if arg is None:
        return None
    try:
        retval = float(arg)
        return retval
    except ValueError:
        return None

# https://tinyurl.com/382t2k8p
def get_argument_value_as_host_port(argument: str, value_separator: str = '=') -> (str, int):
    """
    Get argument value as host & port from command line ignoring string case.

    return      (host: str, port: int)
    host        A string with host ip, otherwise None if argument is of invalid value type.
    port        An int with host port in range 0..65535, otherwise None if argument is of invalid value type.
    """
    arg = get_argument_value_as_str(argument, value_separator)
    try:
        if arg is None:
            return (None, None)
        args = arg.split(':')
        if len(args) != 2:
            return (None, None)
        host = args[0]
        ipaddress.ip_address(host)
        port = int(args[1])
        if port not in range(0, 65535):
            return (None, None)
        return (host, port)
    except ValueError:
        return (None, None)

def get_argument_value_as_int(argument: str, value_separator: str = '=') -> int:
    """
    Get argument value as int from command line ignoring string case.

    return      An int value otherwise None if argument not found or found an invalid value type.
    """
    arg = get_argument_value_as_str(argument, value_separator)
    if arg is None:
        return None
    try:
        retval = int(arg)
        return retval
    except ValueError:
        return None

def get_argument_value_as_str(argument: str, value_separator: str = '=') -> str:
    """
    Get argument value as str from command line ignoring string case.

    return      An int value otherwise None if argument not found.
    """
    argument = argument + value_separator
    argument = argument.lower()
    for arg in sys.argv:
        arg = arg.lower()
        pos = arg.find(argument)
        if pos == 0:
            return arg[len(argument):]
    return None

class TIMECAPS(ctypes.Structure):
    """
    The TIMECAPS structure contains information about the resolution of the timer.

    wPeriodMin  The minimum supported resolution, in milliseconds.
    wPeriodMax  The maximum supported resolution, in milliseconds.
    """
    _fields_ = (
        ('wPeriodMin', wintypes.UINT),
        ('wPeriodMax', wintypes.UINT)
    )

@contextlib.contextmanager
def os_improve_timings(milliseconds=0):
    """Improve Windows internal timers precision using 'with'."""

    # https://realpython.com/python-with-statement/#writing-good-apis-with-context-managers
    # https://stackoverflow.com/questions/38487114/python-2-x-sleep-call-at-millisecond-level-on-windows

    # change windows timing precision
    caps = TIMECAPS()
    ctypes.windll.winmm.timeGetDevCaps(ctypes.byref(caps), ctypes.sizeof(caps))
    milliseconds = min(max(milliseconds, caps.wPeriodMin), caps.wPeriodMax)
    time_begin_period = ctypes.windll.winmm.timeBeginPeriod(milliseconds)
    yield
    ctypes.windll.winmm.timeEndPeriod(milliseconds)

def os_improve_timings_begin(milliseconds=0) -> bool:
    """Improve Windows internal timers precision begin."""

    # https://stackoverflow.com/questions/38487114/python-2-x-sleep-call-at-millisecond-level-on-windows

    # change windows timing precision
    global OS_IMPROVE_TIMINGS_IN_USE
    try:
        caps = TIMECAPS()
        ctypes.windll.winmm.timeGetDevCaps(ctypes.byref(caps), ctypes.sizeof(caps))
        milliseconds = min(max(milliseconds, caps.wPeriodMin), caps.wPeriodMax)
        time_begin_period = ctypes.windll.winmm.timeBeginPeriod(milliseconds)
        OS_IMPROVE_TIMINGS_IN_USE = milliseconds
        return True
    except WindowsError:
        return False

def os_improve_timings_end():
    """Improve Windows internal timers precision end."""

    # https://stackoverflow.com/questions/38487114/python-2-x-sleep-call-at-millisecond-level-on-windows
    try:
        if not OS_IMPROVE_TIMINGS_IN_USE is None:
            ctypes.windll.winmm.timeEndPeriod(OS_IMPROVE_TIMINGS_IN_USE)
        return True
    except WindowsError:
        return False

#
# == END: public attributes
