#-------------------------------------------------------------------------------
# Name:         qtec_multiple_cuts
#
# Purpose:      Creates a G-Code program for a Wire Cutting CNC starting from
#               the user data entries and and sends resulting program to the
#               calling Control Software using the CNC IPC Client Core Engine.
#
# Note:         Compatible with API client version 1.x
#               1 (on 1.x) means interface contract
#               x (on 1.x) means release version
#
# Note:         API Server must listening on 127.0.0.1:8000 (localhost)
#
# Note:         Script checks for singleton so ONLY one instance of script will
#               be permitted to run in same PC
#
# Requires:     tkinter (installed during python installation)
#               pip install tendo
#               pip install screeninfo
#               pip install PySimpleGUI
#               cnc_ipc_client_core.py
#
# Arguments:    qtec_multiple_cuts.py [monitor_number]
#               where nothing or 0 means primary monitor
#               if monitor number is not available then default monitor is used
#
# Author:       support@rosettacnc.com
#
# Created:      15/01/2022
# Copyright:    RosettaCNC (c) 2016-2022
# Licence:      MIT License
# Coding Style  https://www.python.org/dev/peps/pep-0008/
#-------------------------------------------------------------------------------
import cnc_api_client_core as cnc
from tendo import singleton
import PySimpleGUI as sg
import screeninfo
import sys
import os_utils

# == api server default settings
API_SERVER_HOST_DEF = '127.0.0.1'
API_SERVER_PORT_DEF = 8000

# == ui default settings ===
UI_SCALING      = 2.0
UI_USE_MONITOR  = 0             # 0 = primary monitor (default)
UI_STAY_ON_TOP  = True

# == gcode settings ===
gcode_header    = ['G17 G21 G90 G91.1', 'G00 X0 Y0', 'M03', '#<_feed_discesa>=16.76', '#<_feed_risalita>=416', 'F#<_feed_discesa>', ' ']
gcode_tail      = [' ', 'M05', 'M02']

# == ui settings ===
ui_use_monitor  = 0             # 0 = primary monitor (default)
ui_stay_on_top  = True

# gets required monitor number from arguments
try:
    if len(sys.argv) == 2:
        ui_use_monitor = int(sys.argv[1])
except:
    pass

# generates gcode and sends to API server
def generate_gcode(edit_values):

    # adds gcode line
    def add_line(line):
        api.program_gcode_add_text(line)

    # clears gcode program
    api.program_gcode_clear()

    # extracts data from edit values
    data = [dato for dato in edit_values.values()]

    # adds gcode header
    for line in gcode_header:
        add_line(line)

    # adds gcode for pieces heading
    add_line('; Intestatura')
    add_line('M109 P"Intestatura" Q4')
    add_line('G01 F#<_feed_discesa> Y[0-' + str(data[0]) + ']')
    add_line('IF [EXISTS[#<restarting>] NE 0] THEN11')
    add_line('    IF [#<restarting> EQ 0] THEN10')
    add_line('        M66 P#6500 L3 Q' + str(data[1]))
    add_line('    END10')
    add_line('END11')
    if not data[3] and not data[5] and not data[7] and not data[11]:
        add_line('M200')
    add_line('G01 Y0 F#<_feed_risalita>')

    # adds gcode for pieces
    if data[3]:
        for i in range(int(data[3])):
            add_line('; Taglio 1 - pezzo ' + str(i+1))
            add_line('M109 P"Taglio 1 - pezzo ' + str(i+1) + '/' + str(data[3]) + ' - sp.' + str(data[2]) + ' mm" Q4')
            add_line('G91 X' + str(data[2]))
            add_line('G90')
            add_line('G01 F#<_feed_discesa> Y[0-' + str(data[0]) + ']')
            add_line('IF [EXISTS[#<restarting>] NE 0] THEN11')
            add_line('    IF [#<restarting> EQ 0] THEN10')
            add_line('        M66 P#6500 L3 Q' + str(data[1]))
            add_line('    END10')
            add_line('END11')
            add_line('G01 Y0 F#<_feed_risalita>')
    if data[5]:
        for i in range(int(data[5])):
            add_line('; Taglio 2 - pezzo' + str(i+1))
            add_line('M109 P"Taglio 1 - pezzo ' + str(i+1) + '/' + str(data[5]) + ' - sp.' + str(data[4]) + ' mm" Q4')
            add_line('G91 X' + str(data[4]))
            add_line('G90')
            add_line('G01 F#<_feed_discesa> Y[0-' + str(data[0]) + ']')
            add_line('IF [EXISTS[#<restarting>] NE 0] THEN11')
            add_line('    IF [#<restarting> EQ 0] THEN10')
            add_line('        M66 P#6500 L3 Q' + str(data[1]))
            add_line('    END10')
            add_line('END11')
            add_line('G01 Y0 F#<_feed_risalita>')
    if data[7]:
        for i in range(int(data[7])):
            add_line('; Taglio 3 - pezzo' + str(i+1))
            add_line('M109 P"Taglio 1 - pezzo {i+1}/{data[7]} - sp.' + str(data[6]) + ' mm" Q4')
            add_line('G91 X' + str(data[6]))
            add_line('G90')
            add_line('G01 F#<_feed_discesa> Y[0-' + str(data[0]) + ']')
            add_line('IF [EXISTS[#<restarting>] NE 0] THEN11')
            add_line('    IF [#<restarting> EQ 0] THEN10')
            add_line('        M66 P#6500 L3 Q' + str(data[1]))
            add_line('    END10')
            add_line('END11')
            add_line('G01 Y0 F#<_feed_risalita>')
    if data[9]:
        for i in range(int(data[9])):
            add_line('; Taglio 4 - pezzo' + str(i+1))
            add_line('M109 P"Taglio 1 - pezzo {i+1}/{data[9]} - sp.' + str(data[8]) + ' mm" Q4')
            add_line('G91 X' + str(data[8]))
            add_line('G90')
            add_line('G01 F#<_feed_discesa> Y[0-' + str(data[0]) + ']')
            add_line('IF [EXISTS[#<restarting>] NE 0] THEN11')
            add_line('    IF [#<restarting> EQ 0] THEN10')
            add_line('        M66 P#6500 L3 Q' + str(data[1]))
            add_line('    END10')
            add_line('END11')
            add_line('G01 Y0 F#<_feed_risalita>')
    if data[11]:
        for i in range(int(data[11])):
            add_line('; Taglio 5 - pezzo' + str(i+1))
            add_line('M109 P"Taglio 1 - pezzo {i+1}/{data[11]} - sp.' + str(data[10]) + ' mm" Q4')
            add_line('G91 X' + str(data[10]))
            add_line('G90')
            add_line('G01 F#<_feed_discesa> Y[0-' + str(data[0]) + ']')
            add_line('IF [EXISTS[#<restarting>] NE 0] THEN11')
            add_line('    IF [#<restarting> EQ 0] THEN10')
            add_line('        M66 P#6500 L3 Q' + str(data[1]))
            add_line('    END10')
            add_line('END11')
            add_line('G01 Y0 F#<_feed_risalita>')

    # adds gcode tail
    for line in gcode_tail:
        add_line(line)

# checks if an instance is already running (using tendo)
try:
    me = singleton.SingleInstance()
except:
    # notices only one script instance permitted error and quits
    sg.popup_error(
        'PYTHON SCRIPT CANNOT BE EXECUTED!\n' +
        '\n' +
        'An instance of this Python script is already running !\n' +
        '\n'
        'For each computer, or virtual machine, is allowed to start a single instance of the script.\n' +
        '\n'
        'If you see this message it means that a script instance is already loaded and running.',
        no_titlebar=True, keep_on_top=True)
    sys.exit(-1)

# get api server info from command line
host, port = os_utils.get_argument_value_as_host_port('/api_server', ':')
if host is None or port is None:
    host = API_SERVER_HOST_DEF
    port = API_SERVER_PORT_DEF

# get option stayontop from command line
ui_stay_on_top = UI_STAY_ON_TOP
if os_utils.is_argument_available('/stayontop'):
    ui_stay_on_top = True

# get option monitor from command line
monitor_index = os_utils.get_argument_value_as_int('/monitor')
if monitor_index is None:
    monitor_index = UI_USE_MONITOR

# get scaling from command line
ui_scaling = os_utils.get_argument_value_as_float('/scaling')
if ui_scaling is None:
    ui_scaling = UI_SCALING

# creates, sets and checks api client core
api = cnc.CncAPIClientCore()
if api.connect(host, port) == False:
    # notices api connection error and quits
    sg.popup_error(
        'CANNOT CONNECT WITH API SERVER AT',
        host + ':' + str(port) + '!',
        no_titlebar=True, keep_on_top=True)
    quit()

# sets theme, creates layout and shows window
sg.theme('Dark Amber')
layout = [ [sg.Text('Profontita taglio[mm]', size=(17,1)), sg.In(size=(8,1), key='-PROFONDITA-')],
           [sg.Text('Tempo pausa[s]', size=(17,1)), sg.In(size=(8,1), key='-PAUSA-')],
           [sg.Txt('_'  * 30)],
           [sg.Text('Taglio 1 spessore[mm]', size=(17,1)), sg.In(size=(8,1), key='-TAGLIO1_MM-')],
           [sg.Text('Taglio 1 quantita', size=(17,1)), sg.In(size=(8,1), key='-TAGLIO1_PZ-')],
           [sg.Txt('_'  * 30)],
           [sg.Text('Taglio 2 spessore[mm]', size=(17,1)), sg.In(size=(8,1), key='-TAGLIO2_MM-')],
           [sg.Text('Taglio 2 quantita', size=(17,1)), sg.In(size=(8,1), key='-TAGLIO2_PZ-')],
           [sg.Txt('_'  * 30)],
           [sg.Text('Taglio 3 spessore[mm]', size=(17,1)), sg.In(size=(8,1), key='-TAGLIO3_MM-')],
           [sg.Text('Taglio 3 quantita', size=(17,1)), sg.In(size=(8,1), key='-TAGLIO3_PZ-')],
           [sg.Txt('_'  * 30)],
           [sg.Text('Taglio 4 spessore[mm]', size=(17,1)), sg.In(size=(8,1), key='-TAGLIO4_MM-')],
           [sg.Text('Taglio 4 quantita', size=(17,1)), sg.In(size=(8,1), key='-TAGLIO4_PZ-')],
           [sg.Txt('_'  * 30)],
           [sg.Text('Taglio 5 spessore[mm]', size=(17,1)), sg.In(size=(8,1), key='-TAGLIO5_MM-')],
           [sg.Text('Taglio 5 quantita', size=(17,1)), sg.In(size=(8,1), key='-TAGLIO5_PZ-')],
           [sg.Txt('_'  * 30)],
           [sg.Button('Calcola', size=(20,1), key='btn_generate_gcode'), sg.Button('Esci',key='btn_exit')]]
window = sg.Window(
        'Dati per i tagli multipli',
        layout, 
        finalize=True, 
        keep_on_top=ui_stay_on_top,
        scaling=(ui_scaling * 96 / 72)
        )

# places window in center of choosed monitor (requires finalize=True in window creation to get window.size[])
if ui_use_monitor != 0:
    monitors = screeninfo.get_monitors()
    if ui_use_monitor <= len(monitors):
        x = monitors[ui_use_monitor].x
        y = monitors[ui_use_monitor].y
        w = monitors[ui_use_monitor].width
        h = monitors[ui_use_monitor].height
        wx = int(x + ((w - window.size[0]) / 2))
        wy = int(y + ((h - window.size[1]) / 2))
        window.move(wx, wy)

# evaluates main loop
while True:
    event, values = window.read()

    if event == 'btn_generate_gcode':
        generate_gcode(values)

    elif event == sg.WIN_CLOSED or event == 'btn_exit':
        break

# closes api client
api.close()

# closes window
window.close()