#-------------------------------------------------------------------------------
# Name:         module1
#
# Purpose:      Xxx...
#
# Note          Tested with Python x.x.x
#
# Requires:     Xxx...
#
# Author:       support@rosettacnc.com
#
# Created:      12/05/2022
# Copyright:    RosettaCNC (c) 2016-2022
# Licence:      RosettaCNC License 1.0 (RCNC-1.0)
# Coding Style  https://www.python.org/dev/peps/pep-0008/
#-------------------------------------------------------------------------------
import cnc_api_client_core as cnc
from tendo import singleton
import PySimpleGUI as sg
import screeninfo
import sys
import os_utils

# == api server default settings
API_SERVER_HOST_DEF = '127.0.0.1'
API_SERVER_PORT_DEF = 8000

# == ui default settings ===
UI_SCALING      = 4.0
UI_USE_MONITOR  = 0             # 0 = primary monitor (default)
UI_STAY_ON_TOP  = True
UI_THEME        = 'DarkGrey6'
UI_THEME        = 'Dark Amber'
UI_FONT_01      = ("Klartext Mono", 12)
UI_FONT_02      = ("Klartext Mono", 2)

def main():

    def on_update():
        # update cnc api context
        if not ctx.update():
            # closes main window
            view.close_window()

            # closes api client
            api.close()

            # notices api connection error and quits
            sg.popup_error(
                'LOST CONNECTION WITH API SERVER AT',
                api_server_host + ':' + str(api_server_port) + '!',
                no_titlebar=True)
            quit()

        # defines a shortcuts for context info
        cnc_info = ctx.cnc_info
        axes_info = ctx.axes_info

        # evaluates units mode
        ums = ' mm' if cnc_info.units_mode == cnc.UM_METRIC else ' in'
        umv = ' mm/min' if cnc_info.units_mode == cnc.UM_METRIC else ' in/min'

        # evaluates axis colors depending by axes mask
        axis_color = [
            ['white' if cnc_info.axes_mask & cnc.X_AXIS_MASK else 'gray'],
            ['white' if cnc_info.axes_mask & cnc.Y_AXIS_MASK else 'gray'],
            ['white' if cnc_info.axes_mask & cnc.Z_AXIS_MASK else 'gray'],
        ]

        # evaluates axes colors depending by homing done mask status
        homing_color = [
            ['white' if axes_info.homing_done_mask & cnc.X_AXIS_MASK else 'yellow'],
            ['white' if axes_info.homing_done_mask & cnc.Y_AXIS_MASK else 'yellow'],
            ['white' if axes_info.homing_done_mask & cnc.Z_AXIS_MASK else 'yellow'],
        ]
        if not (cnc_info.axes_mask & cnc.X_AXIS_MASK): homing_color[0] = 'gray'
        if not (cnc_info.axes_mask & cnc.Y_AXIS_MASK): homing_color[1] = 'gray'
        if not (cnc_info.axes_mask & cnc.Z_AXIS_MASK): homing_color[2] = 'gray'

        # updates program positions
        window['txt_x'].update(value="{:11.3f}".format(axes_info.program_position[0]) + ums, text_color=homing_color[0])
        window['txt_y'].update(value="{:11.3f}".format(axes_info.program_position[1]) + ums, text_color=homing_color[1])
        window['txt_z'].update(value="{:11.3f}".format(axes_info.program_position[2]) + ums, text_color=homing_color[2])

        # updates feed and pause
        window['txt_feed'].update(value="{:11.0f}".format(cnc_info.feed_reference) + umv)

    # get api server info from command line
    host, port = os_utils.get_argument_value_as_host_port('/api_server', ':')
    if host is None or port is None:
        host = API_SERVER_HOST_DEF
        port = API_SERVER_PORT_DEF

    # get option stayontop from command line
    ui_stay_on_top = UI_STAY_ON_TOP
    if os_utils.is_argument_available('/stayontop'):
        ui_stay_on_top = True

    # get option monitor from command line
    monitor_index = os_utils.get_argument_value_as_int('/monitor')
    if monitor_index is None:
        monitor_index = UI_USE_MONITOR

    # get scaling from command line
    ui_scaling = os_utils.get_argument_value_as_float('/scaling')
    if ui_scaling is None:
        ui_scaling = UI_SCALING

    # creates, sets and checks api client core
    api = cnc.CncAPIClientCore()
    if api.connect(host, port) == False:
        # notices api connection error and quits
        sg.popup_error(
            'CANNOT CONNECT WITH API SERVER AT',
            host + ':' + str(port) + '!',
            no_titlebar=True
        )
        quit()

    # sets theme, creates layout and shows window
    sg.theme(UI_THEME)
    layout = [ [sg.Text('', size=(7,1), font=UI_FONT_02)],
               [sg.Text('X     ', size=(7,1), font=UI_FONT_01), sg.Text('', size=(16,1), font=UI_FONT_01, key='txt_x')],
               [sg.Text('Y     ', size=(7,1), font=UI_FONT_01), sg.Text('', size=(16,1), font=UI_FONT_01, key='txt_y')],
               [sg.Text('Z     ', size=(7,1), font=UI_FONT_01), sg.Text('', size=(16,1), font=UI_FONT_01, key='txt_z')],
               [sg.HorizontalSeparator()],
               [sg.Text('Feed  ', size=(7,1), font=UI_FONT_01), sg.Text('', size=(16,1), font=UI_FONT_01, key='txt_feed')],
               [sg.HorizontalSeparator()],
               [sg.Text('', size=(7,1), font=UI_FONT_02)],
               [sg.Button('Esci', font=UI_FONT_01, key='btn_exit')]]
    window = sg.Window(
        'MONITOR',
        layout,
        finalize=True,
        keep_on_top=ui_stay_on_top,
        scaling=(ui_scaling * 96 / 72)
    )
    window['btn_exit'].expand(True, True, True)

    # places window in center of choosed monitor (requires finalize=True in window creation to get window.size[])
    if monitor_index != 0:
        monitors = screeninfo.get_monitors()
        if monitor_index <= len(monitors):
            x = monitors[monitor_index].x
            y = monitors[monitor_index].y
            w = monitors[monitor_index].width
            h = monitors[monitor_index].height
            wx = int(x + ((w - window.size[0]) / 2))
            wy = int(y + ((h - window.size[1]) / 2))
            window.move(wx, wy)

    # creates a cnc api info contect
    ctx = cnc.CncAPIInfoContext(api)

    # evaluates main loop
    while True:
        event, values = window.read(timeout=10)
        if event is None or event == '-Exit-':
            break

        # == event on timeout ==
        elif event == '__TIMEOUT__':
            on_update()

        elif event == sg.WIN_CLOSED or event == 'btn_exit':
            break

    # closes api client
    api.close()

    # closes window
    window.close()

if __name__ == '__main__':
    main()
